using UnityEngine;
using UnityEngine.UI;
using System;
using System.Collections;
using System.Collections.Generic;
using SimpleJSON;
using System.Linq;
using System.IO;
using MeshVR;
using Request = MeshVR.AssetLoader.AssetBundleFromFileRequest;
using AssetBundles;
using MVR;

// **********************************
// VAMLive Camera 1.0
// **********************************

namespace VAMLivePlugin
{
    public partial class VAMLiveCamera : MVRScript
    {	
	    // Static string
        public static string PLUGIN_PATH;
        private static string CAM_BUNDLE_PATH = "/vamlive_assets.cambundle";
        public static string INFO_PREFIX = "VAMLive: ";
        public static string EXCEPTION_PREFIX = "VAMLive - Exception caught: ";
        
        // JSONStorables
        protected JSONStorableFloat CamOpt_TextureResX;
        protected JSONStorableFloat CamOpt_TextureResY;
        
        protected JSONStorableFloat CamOpt_Fov;
        protected JSONStorableFloat CamOpt_RatioWidth;
        protected JSONStorableFloat CamOpt_RatioHeight;
        
        // The usual help section
        protected JSONStorableString helpText;
        
        // VAMS External values
        public JSONStorableBool VAMLValue_Ready;
        
        // Global UI components
        protected UIDynamicTextField helpWindow;

        // Game Objects, components, assets and assets lists
        protected GameObject VAMLIVECAMERAGO;
        
        protected Transform _TRVAMLCamera;
        
        protected Camera _VAMLCamera;
        
        // Textures
        private RenderTexture _camRT;
        
        // Dynamic UI
        protected List<UIDynamic> ignoreLayoutUI = new List<UIDynamic>();
        
        // Miscs
        protected float _vamexmTimer = 0f;
        protected bool _vamliveInitialized = false;
        protected bool _isBundleLoaded = false;
        
        // Assetbundle
        private Request bundleRequest = null;
        
        // Security
        protected bool _isSceneLoading { get { return SuperController.singleton.isLoading; } }
        protected bool _isSceneFrozen { get { return SuperController.singleton.freezeAnimation; } }

		public override void Init()
        {			 
            try
            {
                if (containingAtom.type != "Empty")
                {
                    SuperController.LogError(EXCEPTION_PREFIX + "Please add VAMLive Camera to an Empty atom");
                    return;
                }
                
                PLUGIN_PATH = Utils.GetPluginPath(this);
                
                /* *** Temporary UIDynamic *** */
                HSVColor hsvcDefault = HSVColorPicker.RGBToHSV(1f, 1f, 1f);
                UIDynamicTextField tmpTextfield;
                UIDynamicPopup tmpPopup;
                UIDynamicSlider tmpSlider;
                UIDynamicButton tmpButton;
                UIDynamicToggle tmpToggle;
                UIDynamicColorPicker tmpColor;
                UIDynamic tmpSpacer;

                // ******* HELP **********************
                helpText = new JSONStorableString("Help",
	                "<size=30><color=#333>" +
	                "<b>VAMLIVE CAMERA HELP:</b>\n" +
	                "This plugin adds a camera that can displayed on a screen from VAMLive.\n\n" +
	                "<b>TEXTURE:</b>\n" +
	                "Set a texture resolution. The bigger it is, the heavier the impact on the framerate. (especially if you have multiple cameras)\n\n" +
	                "Try to optimize the texture size based on your ratio to reduce the performance draw.\n\n" +
	                "You can also use masked screens to <i>hide</i> the low resolution of the texture upclose.\n\n" +
	                "<b>CAMERA:</b>\n" +
	                "You can change the FOV and ratio of your camera.\n\n" +
	                "To avoid having a distorted image, you need to have the same ratio on your camera and screen.\n\n" +
	                "</color></size>"
                );
				
                helpWindow = CreateTextField(helpText, false);
                helpWindow.height = 1150.0f;
                
                /* ********************************* */
                /* **** RIGHT UI                **** */
                /* ********************************* */
                Utils.CreateStaticDescriptionText(this,"VAMLiveCamera","<color=#000><size=35><b>VAMLive (Camera)</b></size></color>",true,40,TextAnchor.MiddleRight);

                // **** Cam options ****
                tmpTextfield = Utils.CreateStaticDescriptionText(this,"Texture","<color=#000><size=28><b>TEXTURE</b></size></color>\n<size=28>Source camera options for the screen.</size>",true,65,TextAnchor.MiddleRight);

                CamOpt_TextureResX = new JSONStorableFloat("CamOpt_TextureResX", 1280f, 64f, 4096f);
                CamOpt_TextureResX.setCallbackFunction += (val) => { CamOpt_TextureResX.valNoCallback = Mathf.ClosestPowerOfTwo((int)val); OnChangeTextureSize(); };
                tmpSlider = CreateSlider(CamOpt_TextureResX, true);
                tmpSlider.valueFormat = "F0";
                tmpSlider.label = "Texture width";

                CamOpt_TextureResY = new JSONStorableFloat("CamOpt_TextureResY", 720f, 64f, 4096f);
                CamOpt_TextureResY.setCallbackFunction += (val) => { CamOpt_TextureResY.valNoCallback = Mathf.ClosestPowerOfTwo((int)val); OnChangeTextureSize(); };
                tmpSlider = CreateSlider(CamOpt_TextureResY, true);
                tmpSlider.valueFormat = "F0";
                tmpSlider.label = "Texture height";
                
                // **** Camera options ****
                tmpSpacer = CreateSpacer(true);
                Utils.SetupSpacer(tmpSpacer, 20f);

                tmpTextfield = Utils.CreateStaticDescriptionText(this,"Camera","<color=#000><size=28><b>CAMERA</b></size></color>",true,35,TextAnchor.MiddleRight);

                CamOpt_Fov = new JSONStorableFloat("CamOpt_Fov", 60f, 1f, 179f);
                CamOpt_Fov.setCallbackFunction += (val) => { CamOpt_Fov.val = Mathf.Round(val); OnChangeCameraFov(); };
                tmpSlider = CreateSlider(CamOpt_Fov, true);
                tmpSlider.valueFormat = "F0";
                tmpSlider.label = "FOV";
                
                CamOpt_RatioWidth = new JSONStorableFloat("CamOpt_RatioWidth", 16f, 1f, 32f);
                CamOpt_RatioWidth.setCallbackFunction += (val) => { OnChangeCameraAspect(); };
                tmpSlider = CreateSlider(CamOpt_RatioWidth, true);
                tmpSlider.valueFormat = "F0";
                tmpSlider.label = "Ratio (width)";

                CamOpt_RatioHeight = new JSONStorableFloat("CamOpt_RatioHeight", 9f, 1f, 9f);
                CamOpt_RatioHeight.setCallbackFunction += (val) => { OnChangeCameraAspect(); };
                tmpSlider = CreateSlider(CamOpt_RatioHeight, true);
                tmpSlider.valueFormat = "F0";
                tmpSlider.label = "Ratio (height)";

                // ******* CREATING MAIN TRIGGERS/ACTIONS/EXTERNAL VALUES ********
                // They are here to be displayed on TOP of every other JSONStorables
                VAMLValue_Ready = new JSONStorableBool("VAMLValue_Ready", false) {isStorable=false,isRestorable=false};
                RegisterBool(VAMLValue_Ready);
                
                /* ***************************************** */
                /* **** REGISTERING STORABLES/ACTIONS   **** */
                /* ***************************************** */
                RegisterFloat(CamOpt_TextureResX);
                RegisterFloat(CamOpt_TextureResY);
                
                RegisterFloat(CamOpt_Fov);
                RegisterFloat(CamOpt_RatioWidth);
                RegisterFloat(CamOpt_RatioHeight);

                // Loading the assetbundle
				Request request = new AssetLoader.AssetBundleFromFileRequest {path = PLUGIN_PATH + CAM_BUNDLE_PATH, callback = OnBundleLoaded};
				AssetLoader.QueueLoadAssetBundleFromFile(request);
            }
            catch(Exception e)
            {
                SuperController.LogError(EXCEPTION_PREFIX + e);
            }
		}
		
		void Update()
		{
			if (_isSceneLoading || _isSceneFrozen) return;

			if (!_isSceneLoading && _isBundleLoaded && _vamliveInitialized == false)
			{
				_camRT = new RenderTexture(1280, 720, 24);
				_camRT.wrapMode = TextureWrapMode.Repeat;
				_VAMLCamera.targetTexture = _camRT;

				_vamliveInitialized = true;
				VAMLValue_Ready.val = true;
				
				// Initializing object
				OnChangeTextureSize();
				OnChangeCameraFov();
				OnChangeCameraAspect();

				// Sending broadcast to screens (to update the list of cameras)
				SendBroadcastToScreens();
			}
			
			// No update if not ready
			if (!_vamliveInitialized) return;

			// Internal timer
			_vamexmTimer += Time.deltaTime;
		}

		// **************************
		// Functions
		// **************************

		
		// **************************************
		// Callbacks & similar ( internal )
		// **************************************
		protected void OnChangeTextureSize()
		{
			if (!_vamliveInitialized) return;
			
			_camRT.Release();
			_camRT.width = (int)CamOpt_TextureResX.val;
			_camRT.height = (int)CamOpt_TextureResY.val;
			_camRT.Create();
		}

		protected void OnChangeCameraFov()
		{
			if (!_vamliveInitialized) return;
			
			_VAMLCamera.fieldOfView = CamOpt_Fov.val;
		}
		
		protected void OnChangeCameraAspect()
		{
			if (!_vamliveInitialized) return;
			
			_VAMLCamera.aspect = CamOpt_RatioWidth.val / CamOpt_RatioHeight.val;
		}
		
		protected void OnBundleLoaded(Request aRequest) {
			try
			{
				// Reading my UI bundle
				bundleRequest = aRequest;
				GameObject tmp = bundleRequest.assetBundle.LoadAsset<GameObject>("assets/vamlive/prefabs/vamlive_camera.prefab");
				VAMLIVECAMERAGO = Instantiate(tmp);
				
				// Parenting to my current Empty
				VAMLIVECAMERAGO.transform.SetParent(containingAtom.reParentObject.GetChild(0).transform, false);
				VAMLIVECAMERAGO.transform.localPosition = new Vector3(0, 0, 0);
				VAMLIVECAMERAGO.transform.localRotation = Quaternion.Euler(0f,0f,0f);
				
				// Grabbing the camera
				_TRVAMLCamera = Utils.GetChildNamed(VAMLIVECAMERAGO.transform, "screen_camera");
				_VAMLCamera = _TRVAMLCamera.GetComponent<Camera>();
				
				SuperController.singleton.onAtomUIDRenameHandlers += OnAtomRename;
				
				_isBundleLoaded = true;
			}
			catch(Exception e)
			{
				SuperController.LogError(EXCEPTION_PREFIX + e);
			}
		}
		
		private void OnAtomRename(string oldid, string newid)
		{
			SendBroadcastToScreens();
		}
		
		// **************************
		// BROADCASTS SEND/RECEIVE
		// **************************
		public void SendBroadcastToScreens()
		{
			foreach(Atom at in SuperController.singleton.GetAtoms() ) {
				if( at.type == "Empty" ) {
					foreach (string stId in at.GetStorableIDs())
					{
						if (stId.Contains(".VAMLiveScreen"))
						{
							at.BroadcastMessage("RefreshCameras");
						}
					}
				}
			}
		}
		
		// **************************
		// LOCAL TOOLS
		// **************************
		private static void logDebug( string debugText ) {
			SuperController.LogMessage( debugText );
		}
		
		private static void logError( string debugText ) {
			SuperController.LogError( debugText );
		}
		
		protected void ToggleUIDynamicGrp(List<UIDynamic> uiGrp,bool enabled)
		{
			foreach( UIDynamic uidElem in uiGrp )
			{
				ToggleUIDynamic(enabled, uidElem);
			}
		}
		
		protected void ToggleUIDynamic(bool enabled, UIDynamic target)
		{
			if (target == null) return;
			
			LayoutElement tmpLE = target.GetComponent<LayoutElement>();

			if (tmpLE != null)
			{
				if (enabled == true)
				{

					tmpLE.transform.localScale = new Vector3(1, 1, 1);
					if( !ignoreLayoutUI.Contains(target) ) tmpLE.ignoreLayout = false;
				}
				else
				{
					tmpLE.transform.localScale = new Vector3(0, 0, 0);
					tmpLE.ignoreLayout = true;
				}
			}
		}
		
		private void SetupButtonWithLayout(UIDynamicButton target, float newSize)
		{
			ContentSizeFitter tmpCSF = target.button.transform.gameObject.AddComponent<ContentSizeFitter>();
			tmpCSF.horizontalFit = UnityEngine.UI.ContentSizeFitter.FitMode.MinSize;
			LayoutElement tmpLE = target.button.transform.gameObject.GetComponent<LayoutElement>();
			RectTransform tmpRectT = target.button.transform.GetComponent<RectTransform>();
			Rect tmpRect = tmpRectT.rect;
			tmpRectT.pivot = new Vector2(0f, 0.5f);
			tmpLE.minWidth = newSize;
			tmpLE.preferredWidth = newSize;
		}
		
		private void SetupButtonWithoutLayout(UIDynamicButton target, float newSize, Vector2 newPosition)
		{
			ContentSizeFitter tmpCSF = target.button.transform.gameObject.AddComponent<ContentSizeFitter>();
			tmpCSF.horizontalFit = UnityEngine.UI.ContentSizeFitter.FitMode.MinSize;
			LayoutElement tmpLE = target.button.transform.gameObject.GetComponent<LayoutElement>();
			RectTransform tmpRectT = target.button.transform.GetComponent<RectTransform>();
			Rect tmpRect = tmpRectT.rect;
			tmpRectT.pivot = new Vector2(0f, 0.5f);
			tmpLE.minWidth = newSize;
			tmpLE.preferredWidth = newSize;
			tmpLE.ignoreLayout = true;
			tmpRectT.anchoredPosition = newPosition;
			ignoreLayoutUI.Add((UIDynamic) target);
		}

		// **************************
		// Time to cleanup !
		// **************************	
		void OnDisable() {

		}

		void OnDestroy()
		{
			// Cleaning GameObjects and textures
			Destroy(VAMLIVECAMERAGO);
			if (_camRT != null) Destroy(_camRT);
			
			SuperController.singleton.onAtomUIDRenameHandlers -= OnAtomRename;
			
			// Unloading assetbundle
			AssetLoader.DoneWithAssetBundleFromFile(PLUGIN_PATH + CAM_BUNDLE_PATH);
		}
    }
}